/*
 * Copyright (C) 1999-2017 Hartmut Seichter
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef TP_CORE_QUATERNION_HPP
#define TP_CORE_QUATERNION_HPP

#include <pw/core/vector.hpp>
#include <pw/core/matrix.hpp>

namespace pw {

/**
 * simplified quaternion class
 */
template <typename T>
class quaternion {

    static constexpr T _sqrt90 = noexcept(std::sqrt(0.5));

protected:

    vector4<T> _q;

public:

    typedef T value_type;

    quaternion() { *this = identity(); }

    quaternion(const T& x,const T& y,const T& z,const T& w) {
        this->set(x,y,z,w);
    }

    inline
    void set(const T& x,const T& y,const T& z,const T& w) {
        _q.set(x,y,z,w);
    }

    inline void set_x(const T& v) { x() = v; }
    inline void set_y(const T& v) { y() = v; }
    inline void set_z(const T& v) { z() = v; }
    inline void set_w(const T& v) { w() = v; }



    inline const vector4<T> as_vector() const { return _q; }

    inline T& x() { return _q.x(); }
    inline T& y() { return _q.y(); }
    inline T& z() { return _q.z(); }
    inline T& w() { return _q.w(); }

    inline const T& x() const { return _q.x(); }
    inline const T& y() const { return _q.y(); }
    inline const T& z() const { return _q.z(); }
    inline const T& w() const { return _q.w(); }


    //! conversion from a matrix
    inline static const quaternion from_matrix(const matrix<4,4,T> &m);

    //! conversion to a matrix
    const matrix<4,4,T> to_matrix() const;

    //! return identiy quaternion
    static quaternion<T> identity();

    static quaternion<T> rotate_180_degree_around_x(); ///< rotate 180 degree around X axis
    static quaternion<T> rotate_180_degree_around_y(); ///< rotate 180 degree around Y axis
    static quaternion<T> rotate_180_degree_around_z(); ///< rotate 180 degree around Z axis

    static quaternion<T> rotate_90_degree_around_x(bool negative = false);
    static quaternion<T> rotate_90_degree_around_y(bool negative = false);
    static quaternion<T> rotate_90_degree_around_z(bool negative = false);

    template <typename AxisAngleType>
    static const quaternion<T> from_axisangle(const AxisAngleType &aa) {

        using std::sin;
        using std::cos;

        const T sinHalfAngle(sin(aa.angle() * T(0.5) ));

        return quaternion<T>(aa.axis().x() * sinHalfAngle, // x
                             aa.axis().y() * sinHalfAngle, // y
                             aa.axis().z() * sinHalfAngle, // z
                             cos(aa.angle() * 0.5)          // w
                             );

    }
};


template <typename T>
const quaternion<T> quaternion<T>::from_matrix(const matrix<4,4,T> &m) {

    using std::sqrt;

    T wtemp = sqrt(T(1) + m.at(0,0) + m.at(1,1) + m.at(2,2)) / T(2.0);

    const T w4 = T(4.0) * wtemp;
    return quaternion<T>(
                (m.at(2,1) - m.at(1,2)) / w4,
                (m.at(0,2) - m.at(2,0)) / w4,
                (m.at(1,0) - m.at(0,1)) / w4,
                wtemp);
}

template <typename T>
const matrix<4,4,T> quaternion<T>::to_matrix() const {

    matrix<4,4,T> m; m.set_identity();

    T xx      = x() * x();
    T xy      = x() * y();
    T xz      = x() * z();
    T xw      = x() * w();

    T yy      = y() * y();
    T yz      = y() * z();
    T yw      = y() * w();

    T zz      = z() * z();
    T zw      = z() * w();

    m.at(0,0) = 1 - 2 * ( yy + zz );
    m.at(0,1) =     2 * ( xy - zw );
    m.at(0,2) =     2 * ( xz + yw );

    m.at(1,0) =     2 * ( xy + zw );
    m.at(1,1) = 1 - 2 * ( xx + zz );
    m.at(1,2) =     2 * ( yz - xw );

    m.at(2,0) =     2 * ( xz - yw );
    m.at(2,1) =     2 * ( yz + xw );
    m.at(2,2) = 1 - 2 * ( xx + yy );

    return m;
}

template <typename T>
quaternion<T> quaternion<T>::identity()
{
    return quaternion<T>(0,0,0,1);
}

template <typename T>
quaternion<T> quaternion<T>::rotate_180_degree_around_x()
{
    return quaternion<T>(1,0,0,0);
}
template <typename T>
quaternion<T> quaternion<T>::rotate_180_degree_around_y()
{
    return quaternion<T>(0,1,0,0);
}

template <typename T>
quaternion<T> quaternion<T>::rotate_180_degree_around_z()
{
    return quaternion<T>(0,0,1,0);
}

template <typename T>
quaternion<T> quaternion<T>::rotate_90_degree_around_x(bool negative/* = false*/)
{
    return quaternion<T>((negative) ? -_sqrt90 : _sqrt90,0,0,_sqrt90);
}

template <typename T>
quaternion<T> quaternion<T>::rotate_90_degree_around_y(bool negative/* = false*/)
{
    return quaternion<T>(0, (negative) ? -_sqrt90 : _sqrt90,0,_sqrt90);
}

template <typename T>
quaternion<T> quaternion<T>::rotate_90_degree_around_z(bool negative/* = false*/)
{
    return quaternion<T>(0,0,(negative) ? -_sqrt90 : _sqrt90, _sqrt90);
}


//
//
//
typedef quaternion<float> quaternionf;
typedef quaternion<double> quaterniond;

}


#endif
